"""Declares various helper functions related to security."""
import asyncio
import contextlib
import time
import typing


@contextlib.asynccontextmanager
async def constant(
    tmin: typing.Union[float, int]
) -> contextlib.AbstractAsyncContextManager:
    """Return a context manager that ensures that the execution time of the
    wrapped code block is constant.
    """
    t1 = time.time()
    try:
        yield
    finally:
        delta = (time.time() - t1)
        if delta < tmin:
            await asyncio.sleep(tmin - delta)



"""Supplies various helper constructs."""
import asyncio
import functools


class allow_sync:
    """Allows the synchronous execution of an async function."""

    @property
    def loop(self):
        try:
            return asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            return loop

    def __init__(self, func):
        self.__func = func
        self.__loop = None
        functools.update_wrapper(self, func)

    def sync(self, *args, **kwargs):
        return self.loop.run_until_complete(self.__func(*args, **kwargs))

    async def __call__(self, *args, **kwargs):
        return await self.__func(*args, **kwargs)



# pylint: skip-file
import asyncio
import unittest

from .. import meta


class AllowSyncDecoratorTestCase(unittest.TestCase):

    def test_run_async_function_sync(self):
        result = foo, bar, *_ = async_function.sync('bar')
        self.assertEqual(foo, 'foo')
        self.assertEqual(bar, 'bar')


@meta.allow_sync
async def async_function(arg, *args, **kwargs):
    return 'foo', arg, args, kwargs



"""Provides the base functionalities to configure a database connection from
environment variables or configuration files.
"""
import copy
import os

import dsnparse
import yaml

from unimatrix.const import SECDIR
from unimatrix.lib.datastructures import DTO


__all__ = ['parse_environment']


def _parse_bool(value):
    return (value == '1') if value is not None else None


ENVIRONMENT_VARIABLES = [
    ('DB_HOST', 'host', False, str),
    ('DB_PORT', 'port', False, int),
    ('DB_NAME', 'name', False, str),
    ('DB_USERNAME', 'username', False, str),
    ('DB_PASSWORD', 'password', False, str),
    ('DB_CONNECTION_MAX_AGE', 'max_age', False, int),
    ('DB_AUTOCOMMIT', 'autocommit', False, _parse_bool)
]


#: Default values by engine.
DEFAULTS = {
    'postgresql': {
        'host': "localhost",
        'port': 5432,
    },
    'mysql': {
        'host': "localhost",
        'port': 3306,
    },
    'sqlite': {
        'name': ":memory:"
    },
    'mssql': {
        'host': 'localhost',
        'port': 1433,
        'options': {
            'driver': 'ODBC Driver 17 for SQL Server',
            'unicode_results': True,
        }
    }
}


class DataSourceName(dsnparse.ParseResult):

    @property
    def admin(self) -> str:
        """Return the DSN used for an administrative connection (e.g. to
        create a new database if it does not exist yet).
        """
        return self.replace(path='')

    @property
    def name(self) -> str:
        return str.lstrip(self.path or '', '/') or None

    @property
    def test(self) -> str:
        """Return the DSN for a transient testing database connection."""
        if self.scheme == 'sqlite':
            return DataSourceName('sqlite://')
        if self.name is None:
            raise ValueError("Must specify a database name.")
        return self.replace(path=self.generate_test_name())

    @classmethod
    def verify(cls, dsn):
        pass

    @property
    def hostloc(self):
        hostloc = self.hostname
        if self.port:
            hostloc = '{hostloc}:{port}'.format(hostloc=hostloc, port=self.port)
        return hostloc or 'localhost'

    def generate_test_name(self) -> str:
        """Return a string containing a unique name for a transient testing
        database.
        """
        return f'test_{bytes.hex(os.urandom(3))}'

    def replace(self, **kwargs):
        """Replace the parameters specified in the keyword arguments and
        return a new instance.
        """
        scheme = kwargs.setdefault('scheme', self.scheme)
        hostname = kwargs.setdefault('hostname', self.hostname)
        params = self.parse(self.geturl())
        params.pop('dsn')
        return dsnparse.parse(
            f'{scheme}://{hostname}',
            parse_class=type(self), **{**params, **kwargs})


def encode_dsn(opts: DTO) -> str:
    """Encodes the database configuration into a Data Source Name (DSN)."""
    dsn = dsnparse.parse(f'{opts.engine}://', DataSourceName)
    if opts.engine == 'sqlite':
        if opts.host is None:
            dsn.setdefault('hostname', '/')
        if opts.get('name') not in (':mem:', ':memory:', None):
            dsn.setdefault('hostname', opts.name)
    else:
        dsn.setdefault('username', opts.get('username') or '')
        dsn.setdefault('password', opts.get('password') or '')
        dsn.setdefault('path', opts.get('name') or '')
        if opts.host is not None:
            dsn.setdefault('hostname', opts.host)

    if opts.engine == 'postgresql':
        dsn.setdefault('port', opts.get('port') or 5432)
    elif opts.engine == 'mssql':
        dsn.setdefault('port', opts.get('port') or 1433)
    elif opts.engine == 'mysql':
        dsn.setdefault('port', opts.get('port') or 3306)
    elif opts.engine == 'sqlite':
        pass
    else:
        raise NotImplementedError(opts.engine)
    if opts.engine != 'sqlite':
        assert dsn.hostname is not None
    return dsn


def parse_environment(env):
    """Parses a database connection from the operating system
    environment based on the ``DB_*`` variables. Return a
    datastructure containing the configuration, or ``None`` if
    it was not defined.
    """
    # If no engine is specified, then the application does not
    # get its database connection configuration from environment
    # variables.
    engine = env.get('DB_ENGINE')
    if engine is None:
        return None

    if engine not in DEFAULTS:
        raise ValueError("Unsupported DB_ENGINE: %s" % engine)

    opts = DTO(engine=engine, options={})
    for name, attname, required, cls in ENVIRONMENT_VARIABLES:
        default = DEFAULTS[engine].get(attname)
        value = env.get(name)
        if value is None:
            value = default
        opts[attname] = cls(value) if value is not None else value
        assert not required #nosec

        if DEFAULTS[engine].get('options'):
            opts['options'] = copy.deepcopy(DEFAULTS[engine]['options'])

        # Engine specific - FIXME
        if engine == 'mssql':
            if env.get('DB_DRIVER'):
                opts['options']['driver'] = env['DB_DRIVER']

        opts['dsn'] = encode_dsn(opts)

    opts.engine = engine
    return opts.as_dto()


def load(env=None, config_dir=None):
    """Loads the configured database connections from the operating
    system environment variables and the Unimatrix configuration files.

    Database connections are loaded with the following precedence:

    - Environment variables
    - Unimatrix Database Connection (UDC) files.

    Args:
        env (dict): a dictionary holding environment variables. Defaults
            to ``os.environ``.
        config_dir (str): a directory on the local filesystem holding
            the database connection configuration files. Defaults to
            `SECDIR/rdbms/connections`.

    Returns:
        dict
    """
    env = copy.deepcopy(env or os.environ)
    config_dir = config_dir or os.path.join(SECDIR, 'rdbms/connections')
    default_connection = env.get('DB_DEFAULT_CONNECTION')
    use_environment = not bool(os.listdir(config_dir))\
        if os.path.exists(config_dir) else True
    if not use_environment and not default_connection:
        raise ValueError("Set the DB_DEFAULT_CONNECTION environment variable.")
    connections = {}
    if use_environment:
        self = parse_environment(env)
        if self:
            connections['self'] = self
    else:
        for fn in os.listdir(config_dir):
            connections[fn] = yaml.safe_load(open(os.path.join(config_dir, fn)))

        if default_connection not in connections:
            raise LookupError(
                "Default connection does not exist: " + default_connection)

        connections['self'] = connections[default_connection]

    return connections



# pylint: skip-file
import os
import unittest

import unimatrix.runtime

from . import mock_boot_module as mocks



class BootModuleTestCase(unittest.TestCase):

    def setUp(self):
        if hasattr(mocks, 'func'):
            delattr(mocks, 'func')
        os.environ.pop('UNIMATRIX_BOOT_MODULE', None)

    @unittest.expectedFailure
    def test_sync_callable(self):
        os.environ['UNIMATRIX_BOOT_MODULE'] = 'unimatrix.runtime.tests.mock_boot_module.sync_func'
        args, kwargs = unimatrix.runtime.on.sync('boot', 1, foo=1)
        self.assertIn(1, args)
        self.assertIn('foo', kwargs)

    @unittest.expectedFailure
    def test_callable(self):
        os.environ['UNIMATRIX_BOOT_MODULE'] = 'unimatrix.runtime.tests.mock_boot_module.func'
        args, kwargs = unimatrix.runtime.on.sync('boot', 1, foo=1)
        self.assertIn(1, args)
        self.assertIn('foo', kwargs)

    def test_module_boot(self):
        os.environ['UNIMATRIX_BOOT_MODULE'] = 'unimatrix.runtime.tests.mock_boot_module'
        args, kwargs = unimatrix.runtime.on.sync('boot', 1, foo=1)
        self.assertIn(1, args)
        self.assertIn('foo', kwargs)

    def test_module_shutdown(self):
        os.environ['UNIMATRIX_BOOT_MODULE'] = 'unimatrix.runtime.tests.mock_boot_module'
        args, kwargs = unimatrix.runtime.on.sync('shutdown', 1, foo=1)
        self.assertIn(1, args)
        self.assertIn('foo', kwargs)



"""Exposes functions related to HTTP."""
import urllib.parse

from unimatrix.lib.datastructures import DTO


def parse_qs(value):
    """Parses a query string into a dictionary"""
    params = {}
    for key, value in urllib.parse.parse_qsl(value):
        if key in params:
            current = params.pop(key)
            if isinstance(current, str):
                current = [current]
            current.append(value)
            params[key] = current
            continue

        params[key] = value

    return DTO.fromdict(params)



"""Exposes the settings declared in the environment variable
``UNIMATRIX_SETTINGS_MODULE``."""
import contextlib
import functools
import importlib
import inspect
import itertools
import os

import unimatrix.const
import unimatrix.environ
from unimatrix.exceptions import ImproperlyConfigured


class Settings:
    """Proxy object to load settings lazily."""
    not_initialized = object()

    @property
    def available(self) -> list:
        """Return the list of available settings."""
        if not self.__is_initialized:
            self.initialize()
        return list(sorted(set(itertools.chain(
            self.__overrides\
                if self.__overrides != self.not_initialized\
                else [],
            dict.keys(self.__values),
            dict.keys(self.__environ)
        ))))

    def __init__(self, env=None):
        self.__is_initialized = False
        self.__environ = env or {}
        self.__overrides = self.not_initialized
        self.__values = {}

    def destroy(self):
        """Reset the :class:`~unimatrix.conf.Settings` object to
        a pristine state.
        """
        self.__is_initialized = False
        self.__environ = {}
        self.__overrides = self.not_initialized
        self.__values = {}

    def initialize(self):
        """Initialize the settings using system environment variables and
        the Python module specified by ``UNIMATRIX_SETTINGS_MODULE``.
        """
        if self.__is_initialized:
            raise RuntimeError
        settings = self._get_settings_module()
        if settings is not None:
            for attname, value in inspect.getmembers(settings):
                if not str.isupper(attname):
                    continue
                self.__values[attname] = value

        # Add the values to the environ dictionary.
        for attname, value in inspect.getmembers(unimatrix.environ):
            if not str.isupper(attname):
                continue
            self.__environ[attname] = value

        self.__is_initialized = True

    @contextlib.contextmanager
    def override(self, **settings):
        """Override the settings with the given values. Not thread-safe."""
        if self.__overrides != self.not_initialized:
            raise RuntimeError("Nested overrides are not supported.")
        try:
            self.__overrides = settings
            yield
        finally:
            self.__overrides = self.not_initialized

    def with_overrides(self, **overrides):
        """Decorates a function to use the given settings overrides."""
        def decorator_factory(func):
            @functools.wraps(func)
            def f(*args, **kwargs):
                with self.override(**overrides):
                    return func(*args, **kwargs)
            return f
        return decorator_factory

    def _get_settings_module(self):
        os.environ.setdefault('DEPLOYMENT_ENV', 'production')
        settings = None
        if os.getenv('UNIMATRIX_SETTINGS_MODULE'):
            module_qualname = os.getenv('UNIMATRIX_SETTINGS_MODULE')
            try:
                settings = importlib.import_module(module_qualname)
            except ImportError:
                raise ImproperlyConfigured(
                    "Can not import settings module %s" % module_qualname)
        return settings

    def __getattr__(self, attname):
        if not self.__is_initialized:
            self.initialize()
        value = sentinel = object()
        if self.__overrides != self.not_initialized:
            value = self.__overrides.get(attname, sentinel)
        if value == sentinel:
            value = self.__values.get(attname, sentinel)
        if value == sentinel:
            value = self.__environ.get(attname, sentinel)
        if value == sentinel:
            raise AttributeError("No such setting: %s" % attname)
        return value


settings = Settings()
